#define NDEBUG
#include <omp.h>
#include <sdsl/suffixarrays.hpp>
#include <queue>

using namespace sdsl;
using namespace std;

typedef bit_vector::size_type size_type;
typedef wt_huff<bit_vector, rank_support_v<>, select_support_bs< rank_support_v5<> >, select_support_bs< rank_support_v5<> > > tWTH;

const uint32_t UNDEF = -1;

struct interval
{
	unsigned char c;
	size_type lb;
	size_type rb;
	interval(unsigned char _c=0, size_type _lb=0, size_type _rb=0) : c(_c), lb(_lb), rb(_rb) { }
};

bool operator<(interval const& a, interval const& b)
{
	return a.c < b.c;
}

struct recursive_call
{
	size_type k;
	size_type lb;
	size_type rb;
	size_type l;
	recursive_call(size_type _k=0, size_type _lb=0, size_type _rb=0, size_type _l=0) : k(_k), lb(_lb), rb(_rb), l(_l) { }
};

void fill_reduzible_lcp_values(int_vector_file_buffer<8> &bwt_buf, int_vector<32> &lcpr, vector<size_type> &C)
{
	stack<size_type> st;
	size_type N=lcpr.size();

	// Init BWT
	bwt_buf.reset();
	
	// Write LF values into LCP-array
	vector<size_type> LFcounter(C.size(), 0);
	bit_vector todo(N, 0);
	for(size_type j=0; j<C.size(); ++j)
	{
		LFcounter[j] = C[j];
	}
	for(size_type i=0, r_sum=0, r=0; r_sum < N; )
	{
		// proccess next r values
		for( ; i < r_sum+r; ++i)
		{
			unsigned char c = bwt_buf[i-r_sum];

			// write LF value
			if(lcpr[i] == UNDEF)
			{
				todo[i] = 1;
				lcpr[i] = LFcounter[c];
			}

			// update LF value
			++LFcounter[c];
		}

		// Count how many were proccessed and how many values will be proccess next
		r_sum += r;
		r = bwt_buf.load_next_block();
	}

	for(size_type i=0; i<N; ++i)
	{
		size_type k=i;
		while(todo[k])
		{
			st.push(k);
			k = lcpr[k];
		}
		size_type l = lcpr[k];
		while(!st.empty())
		{
			--l;
			lcpr[st.top()] = l;
			todo[st.top()] = 0;
			st.pop();
		}
	}
	return;
}

void bwt_and_lcp_rev(size_type k, size_type i, size_type j, size_type l, int_vector<8> &text, int_vector<> &sa, tWTH &wt_bwt, int_vector<8> &bwtr, int_vector<32> &lcpr, vector<size_type> &C)
{
	size_type N = text.size();						// Input length
	size_type realsigma = wt_bwt.sigma;				// Alphabet size
	vector<interval> intervals(realsigma);			// list of intervals

	size_type quantity;								// quantity of characters in interval
	vector<unsigned char> pos2char(realsigma);		// list of characters in the interval
	vector<size_type> resi(realsigma);				// number of occurrence of character in [0 .. i-1]
	vector<size_type> resj(realsigma);				// number of occurrence of character in [0 .. j-1]

	stack<recursive_call> s;
	s.push(recursive_call(k, i, j, l));

	while(!s.empty())
	{
		k = s.top().k;
		i = s.top().lb;
		j = s.top().rb;
		l = s.top().l;
		s.pop();

		wt_bwt.interval_symbols(i, j, quantity, pos2char, resi, resj);
		while(quantity == 1)
		{
			unsigned char c = pos2char[0];
			i = C[c] + resi[0];
			j = C[c] + resj[0];
			++l;
			wt_bwt.interval_symbols(i, j, quantity, pos2char, resi, resj);
		}
		
		for(size_type idx=0; idx<quantity; ++idx)
		{
			unsigned char c = pos2char[idx];
			size_type lb = C[c] + resi[idx];
			size_type rb = C[c] + resj[idx];
			intervals[idx] = interval(c, lb, rb);
		}
		sort(intervals.begin(), intervals.begin()+quantity);

		for(size_type idx=0; idx<quantity; ++idx)
		{
			size_type lb = intervals[idx].lb;
			size_type rb = intervals[idx].rb;

			size_type text_position1 = sa[lb] + l + 1;

			if(lb == rb-1) // rb is exlusive
			{
				if(text_position1 >= N)
				{
					text_position1 -= N;
				}
				bwtr[k] = text[text_position1];
				++k;
			}
			else
			{
				size_type text_position2 = sa[rb-1]+l+1;

				unsigned char textchar_1 = text[text_position1];
				unsigned char textchar_2 = text[text_position2];
				if(textchar_1 == textchar_2)
				{
					for(size_type q=k; q < k+rb-lb; ++q)
					{
						bwtr[q] = text[text_position1];
					}
				}
				else if(lb+2 == rb) // This optimization couldn't be explained in the paper due to space reasons
				{
					text_position1 -= (l+1);
					text_position2 -= (l+1);
					size_type ll = l;
					while(text[text_position1] == text[text_position2])
					{
						--text_position1;
						--text_position2;
						++ll;
					}

					if(text[text_position1] < text[text_position2])
					{
						bwtr[k] = textchar_1;
						bwtr[k+1] = textchar_2;
					}
					else
					{
						bwtr[k] = textchar_2;
						bwtr[k+1] = textchar_1;
					}
					lcpr[k+1] = ll;
				}
				else
				{
					s.push(recursive_call(k, lb, rb, l+1));
				}
				k = k + rb - lb;
			}

			if( idx+1<quantity )
			{
				lcpr[k] = l;
			}
		}
	}
	return;
}

void prepare(size_type k, size_type i, size_type j, size_type l, int_vector<8> &text, int_vector<> &sa, tWTH &wt_bwt, int_vector<8> &bwtr, int_vector<32> &lcpr, vector<size_type> &C, vector<recursive_call> &ausgabe, size_type P)
{
	size_type N = text.size();						// Input length
	size_type realsigma = wt_bwt.sigma;				// Alphabet size
	vector<interval> intervals(realsigma);			// list of intervals

	size_type quantity;								// quantity of characters in interval
	vector<unsigned char> pos2char(realsigma);		// list of characters in the interval
	vector<size_type> resi(realsigma);				// number of occurrence of character in [0 .. i-1]
	vector<size_type> resj(realsigma);				// number of occurrence of character in [0 .. j-1]

	queue<recursive_call> q;
	q.push(recursive_call(k, i, j, l));

	while(!q.empty() and q.size() < P)
	{
		k = q.front().k;
		i = q.front().lb;
		j = q.front().rb;
		l = q.front().l;
		q.pop();

		wt_bwt.interval_symbols(i, j, quantity, pos2char, resi, resj);

		for(size_type idx=0; idx<quantity; ++idx)
		{
			unsigned char c = pos2char[idx];
			size_type lb = C[c] + resi[idx];
			size_type rb = C[c] + resj[idx];
			intervals[idx] = interval(c, lb, rb);
		}
		sort(intervals.begin(), intervals.begin()+quantity);

		for(size_type idx=0; idx<quantity; ++idx)
		{
			size_type lb = intervals[idx].lb;
			size_type rb = intervals[idx].rb;
			size_type text_position1 = sa[lb] + l + 1;

			if(lb == rb-1) // rb is exlusive
			{
				if(text_position1 >= N)
				{
					text_position1 -= N;
				}
				bwtr[k] = text[text_position1];
				++k;
			}
			else
			{
				size_type text_position2 = sa[rb-1]+l+1;

				unsigned char textchar_1 = text[text_position1];
				unsigned char textchar_2 = text[text_position2];
				if(textchar_1 == textchar_2)
				{
					for(size_type q=k; q < k+rb-lb; ++q)
					{
						bwtr[q] = text[text_position1];
					}
				}
				else if(lb+2 == rb) // This optimization couldn't be explained in the paper due to space reasons
				{
					text_position1 -= (l+1);
					text_position2 -= (l+1);
					size_type ll = l;
					while(text[text_position1] == text[text_position2])
					{
						--text_position1;
						--text_position2;
						++ll;
					}

					if(text[text_position1] < text[text_position2])
					{
						bwtr[k] = textchar_1;
						bwtr[k+1] = textchar_2;
					}
					else
					{
						bwtr[k] = textchar_2;
						bwtr[k+1] = textchar_1;
					}
					lcpr[k+1] = ll;
				}
				else
				{
					q.push(recursive_call(k, lb, rb, l+1));
				}
				k = k + rb - lb;
			}

			if( idx+1<quantity )
			{
				lcpr[k] = l;
			}
		}
	}

	while(!q.empty())
	{
		ausgabe.push_back(q.front());
		q.pop();
	}
	return;
}

int main(int argc, char **argv)
{
	tMSS file_map;

	if(argc!=6)
	{
		cout << "Program takes path of the Text, the SA, the BWT, a FILENAME and a NUMBER. It creates (with at most NUMBER CPUs) the BWT and LCP array of reverse string and stores it into bwtr_FILENAME and lcpr_FILENAME." << endl;
		cout << "\tPath to Input text" << endl;
		cout << "\tPath to Suffix Array" << endl;
		cout << "\tPath to BWT" << endl;
		cout << "\tFILENAME" << endl;
		cout << "\tNUMBER of CPUs" << endl;
		cout << endl;
		cout << "\tOutputfile: bwtr_FILENAME, lcpr_FILENAME" << endl;
		return 1;
	}
	file_map["text"] = argv[1];
	file_map["sa"] = argv[2];
	file_map["bwt"] = argv[3];
	file_map["bwtr"] = "bwtr_" + util::to_string(argv[4]);
	file_map["lcpr"] = "lcpr_" + util::to_string(argv[4]);
	file_map["lf"] = "lfr_" + util::to_string(argv[4]);
	size_type max_cpus = atol(argv[5]);

	size_type buffer_size = 10000000;

	size_type quantity;							// quantity of characters in interval
	vector<unsigned char> pos2char(256);		// list of characters in the interval
	vector<size_type> resi(256);				// number of occurrence of character in [0 .. i-1]
	vector<size_type> resj(256);				// number of occurrence of character in [0 .. j-1]

	// Load Text, SA, BWT and create Wavelet Tree of the BWT
	write_R_output("init","load txt, sa & bwt","begin", 0, 0);
	int_vector<8> text;
	util::load_from_file(text, file_map["text"].c_str() );
	int_vector<> sa;
	util::load_from_file(sa, file_map["sa"].c_str() );
	int_vector_file_buffer<8> bwt_buf(file_map["bwt"].c_str(), buffer_size);
	size_type N = bwt_buf.int_vector_size;		// Input length
	tWTH wt_bwt(bwt_buf, N);

	// Create C-array
	vector<size_type> C(256, 0);				// C-Array: C[i] = number of occurrences of characters < i in the input
	wt_bwt.interval_symbols(0, N, quantity, pos2char, resi, resj);
	for(size_type i=0; i<quantity; ++i)
	{
		C[pos2char[i]+1] = resj[i];
	}
	for(size_type i=1; i<C.size()-1; ++i)
	{
		C[i+1] += C[i];
	}
	write_R_output("init","load txt, sa & bwt","end  ", 0, 0);

	write_R_output("bwt","const. bwtr&lcpr ","begin", 0, 0);
	vector<recursive_call> ausgabe;
	int_vector<8> bwtr(N);
	int_vector<32> lcpr(N, UNDEF);
	size_type P = 1000;

	write_R_output("bwt","expand_list      ","begin", 0, 0);
	prepare(0, 0, N, 0, text, sa, wt_bwt, bwtr, lcpr, C, ausgabe, P);
	write_R_output("bwt","expand_list      ","end  ", 0, 0);
	size_type k, i, j, l;

	omp_set_num_threads(max_cpus);
	#pragma omp parallel for private(k,i,j,l) schedule (dynamic,1)
	for(size_type ii=0; ii<ausgabe.size(); ++ii)
	{
		k = ausgabe[ii].k;
		i = ausgabe[ii].lb;
		j = ausgabe[ii].rb;
		l = ausgabe[ii].l;
		bwt_and_lcp_rev(k, i, j, l, text, sa, wt_bwt, bwtr, lcpr, C);
	}
	lcpr[0] = 0;
	write_R_output("bwt","const. bwtr&lcpr ","end  ", 0, 0);

	write_R_output("bwt","store bwtr       ","begin", 0, 0);
	util::store_to_file(bwtr, file_map["bwtr"].c_str());
	write_R_output("bwt","store bwtr       ","end", 0, 0);

	write_R_output("lcp","fill reduzible   ","begin", 0, 0);
	int_vector_file_buffer<8> bwtr_buf(file_map["bwtr"].c_str(), buffer_size);
	fill_reduzible_lcp_values(bwtr_buf, lcpr, C);
	write_R_output("lcp","fill reduzible   ","end", 0, 0);

	write_R_output("lcp","store lcpr       ","begin", 0, 0);
	util::store_to_file(lcpr, file_map["lcpr"].c_str());
	write_R_output("lcp","store lcpr       ","end", 0, 0);

	return 0;
}

